# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes to represent config sections."""

from __future__ import annotations
from typing import Literal, Optional

from pydantic import field_validator, Field
from pydantic.dataclasses import dataclass

import neuron

import bluecellulab

# libsonata reorganized it's module layout; maintain compatibility with both:
# https://github.com/BlueBrain/libsonata/pull/345
try:
    from libsonata._libsonata import Conditions as LibSonataConditions
except ImportError:
    from libsonata._libsonata import SimulationConfig
    LibSonataConditions = SimulationConfig.Conditions


def string_to_bool(value: str) -> bool:
    """Convert a string to a boolean."""
    if value.lower() in ("true", "1"):
        return True
    if value.lower() in ("false", "0"):
        return False
    raise ValueError(f"Invalid boolean value {value}")


@dataclass(frozen=True, config=dict(extra="forbid"))
class ConditionEntry:
    """For mechanism specific conditions."""
    minis_single_vesicle: Optional[int] = Field(None, ge=0, le=1)
    init_depleted: Optional[int] = Field(None, ge=0, le=1)


@dataclass(frozen=True, config=dict(extra="forbid"))
class MechanismConditions:
    """For mechanism specific conditions."""
    ampanmda: Optional[ConditionEntry] = None
    gabaab: Optional[ConditionEntry] = None
    glusynapse: Optional[ConditionEntry] = None


@dataclass(frozen=True, config=dict(extra="forbid"))
class Conditions:
    mech_conditions: Optional[MechanismConditions] = None
    celsius: Optional[float] = None
    v_init: Optional[float] = None
    extracellular_calcium: Optional[float] = None
    randomize_gaba_rise_time: Optional[bool] = None

    @classmethod
    def from_blueconfig(cls, condition_entries: dict) -> Conditions:
        """Create a Conditions instance from a blueconfig section."""
        randomize_gaba_risetime = condition_entries.get("randomize_Gaba_risetime")
        if randomize_gaba_risetime is not None:
            randomize_gaba_risetime = string_to_bool(randomize_gaba_risetime)

        msv_val = condition_entries.get("SYNAPSES__minis_single_vesicle", None)
        init_dep_val = condition_entries.get("SYNAPSES__init_depleted", None)
        mech_conditions = MechanismConditions(
            ampanmda=ConditionEntry(msv_val, init_dep_val),
            gabaab=ConditionEntry(msv_val, init_dep_val),
            glusynapse=ConditionEntry(msv_val, init_dep_val),
        )
        return cls(
            mech_conditions=mech_conditions,
            extracellular_calcium=condition_entries.get("cao_CR_GluSynapse", None),
            randomize_gaba_rise_time=randomize_gaba_risetime,
        )

    @classmethod
    def from_sonata(cls, condition_entries: LibSonataConditions) -> Conditions:
        """Create a Conditions instance from a SONATA section."""
        # sonata stores it as bool, hoc needs it as int
        msv_ampanmda = msv_gabaab = msv_glusynapse = None
        init_dep_ampanmda = init_dep_gabaab = init_dep_glusynapse = None
        mech_dict = condition_entries.mechanisms
        if mech_dict is not None:
            ampanmda = mech_dict.get("ProbAMPANMDA_EMS", None)
            if ampanmda is not None:
                msv_ampanmda = ampanmda.get("minis_single_vesicle", None)
                init_dep_ampanmda = ampanmda.get("init_depleted", None)
            gabaab = mech_dict.get("ProbGABAAB_EMS", None)
            if gabaab is not None:
                msv_gabaab = gabaab.get("minis_single_vesicle", None)
                init_dep_gabaab = gabaab.get("init_depleted", None)
            glusynapse = mech_dict.get("GluSynapse", None)
            if glusynapse is not None:
                msv_glusynapse = glusynapse.get("minis_single_vesicle", None)
                init_dep_glusynapse = glusynapse.get("init_depleted", None)

        mech_conditions = MechanismConditions(
            ampanmda=ConditionEntry(msv_ampanmda, init_dep_ampanmda),
            gabaab=ConditionEntry(msv_gabaab, init_dep_gabaab),
            glusynapse=ConditionEntry(msv_glusynapse, init_dep_glusynapse),
        )

        return cls(
            mech_conditions=mech_conditions,
            celsius=condition_entries.celsius,
            v_init=condition_entries.v_init,
            extracellular_calcium=condition_entries.extracellular_calcium,
            randomize_gaba_rise_time=condition_entries.randomize_gaba_rise_time,
        )

    @classmethod
    def init_empty(cls) -> Conditions:
        """Create an empty conditions object to be used when no condition is
        specified."""
        return cls(
            mech_conditions=None,
            celsius=None,
            v_init=None,
            extracellular_calcium=None,
            randomize_gaba_rise_time=None,
        )


@dataclass(frozen=True, config=dict(extra="forbid"))
class ConnectionOverrides:
    source: str
    target: str
    delay: Optional[float] = None
    weight: Optional[float] = None
    spont_minis: Optional[float] = None
    synapse_configure: Optional[str] = None
    mod_override: Optional[Literal["GluSynapse"]] = None

    @field_validator("mod_override")
    @classmethod
    def validate_mod_override(cls, value):
        """Make sure the mod file to override is present."""
        if isinstance(value, str) and not hasattr(neuron.h, value):
            raise bluecellulab.ConfigError(
                f"Mod file for {value} is not found.")
        return value

    @classmethod
    def from_blueconfig(cls, conn_entry: dict) -> ConnectionOverrides:
        """Create a ConnectionOverrides instance from a blueconfig section."""
        return cls(
            source=conn_entry["Source"],
            target=conn_entry["Destination"],
            delay=conn_entry.get("Delay", None),
            weight=conn_entry.get("Weight", None),
            spont_minis=conn_entry.get("SpontMinis", None),
            synapse_configure=conn_entry.get("SynapseConfigure", None),
            mod_override=conn_entry.get("ModOverride", None),
        )

    @classmethod
    def from_sonata(cls, conn_entry: dict) -> ConnectionOverrides:
        """Create a ConnectionOverrides instance from a SONATA section."""
        return cls(
            source=conn_entry["source"],
            target=conn_entry["target"],
            delay=conn_entry.get("delay", None),
            weight=conn_entry.get("weight", None),
            spont_minis=conn_entry.get("spont_minis", None),
            synapse_configure=conn_entry.get("synapse_configure", None),
            mod_override=conn_entry.get("mod_override", None),
        )



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Main importer of bluecellulab."""

import importlib_resources as resources
import logging
import os
from types import ModuleType

import neuron

from bluecellulab.exceptions import BluecellulabError
from bluecellulab.utils import CaptureOutput, run_once


logger = logging.getLogger(__name__)


def import_mod_lib(neuron: ModuleType) -> str:
    """Import mod files."""
    res = ""
    if 'BLUECELLULAB_MOD_LIBRARY_PATH' in os.environ:
        # Check if the current directory contains 'x86_64'.
        if os.path.isdir('x86_64'):
            raise BluecellulabError("BLUECELLULAB_MOD_LIBRARY_PATH is set"
                                    " and current directory contains the x86_64 folder."
                                    " Please remove one of them.")

        mod_lib_path = os.environ["BLUECELLULAB_MOD_LIBRARY_PATH"]
        if mod_lib_path.endswith(".so"):
            neuron.h.nrn_load_dll(mod_lib_path)
        else:
            neuron.load_mechanisms(mod_lib_path)
        res = mod_lib_path
    elif os.path.isdir('x86_64'):
        # NEURON 8.* automatically load these mechamisms
        res = os.path.abspath('x86_64')
    else:
        res = "No mechanisms are loaded."

    return res


def import_hoc(neuron: ModuleType) -> None:
    """Import hoc dependencies."""
    neuron.h.load_file("stdrun.hoc")  # nrn
    hoc_files = [
        "Cell.hoc",  # ND
        "TDistFunc.hoc",  # ND, test dependency
        "TStim.hoc",  # ND
    ]

    for hoc_file in hoc_files:
        hoc_file_path = str(resources.files("bluecellulab") / f"hoc/{hoc_file}")
        with CaptureOutput() as stdoud:
            neuron.h.load_file(hoc_file_path)
            logger.debug(f"Loaded {hoc_file}. stdout from the hoc: {stdoud}")


def print_header(neuron: ModuleType, mod_lib_path: str) -> None:
    """Print bluecellulab header to stdout."""
    logger.debug(f"Imported NEURON from: {neuron.__file__}")
    logger.debug(f"Mod lib: {mod_lib_path}")


@run_once
def _load_mod_files() -> None:
    """Import hoc and mod files."""
    logger.debug("Loading the mod files.")
    mod_lib_paths = import_mod_lib(neuron)
    print_header(neuron, mod_lib_paths)


def load_mod_files(func):
    def wrapper(*args, **kwargs):
        _load_mod_files()
        return func(*args, **kwargs)
    return wrapper



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Factory that separates Synapse creation from Synapse instances."""
from __future__ import annotations  # PEP-563 forward reference annotations
from enum import Enum
import logging
import math

import neuron
import numpy as np

import pandas as pd

import bluecellulab
from bluecellulab.exceptions import BluecellulabError
from bluecellulab.synapse import Synapse, GabaabSynapse, AmpanmdaSynapse, GluSynapse
from bluecellulab.circuit.config.sections import Conditions
from bluecellulab.circuit.synapse_properties import SynapseProperties, SynapseProperty
from bluecellulab.synapse.synapse_types import SynapseHocArgs
from bluecellulab.type_aliases import NeuronSection


SynapseType = Enum("SynapseType", "GABAAB AMPANMDA GLUSYNAPSE")

logger = logging.getLogger(__name__)


class SynapseFactory:
    """Creates different types of Synapses."""

    @classmethod
    def create_synapse(
        cls,
        cell: bluecellulab.Cell,
        syn_id: tuple[str, int],
        syn_description: pd.Series,
        condition_parameters: Conditions,
        popids: tuple[int, int],
        extracellular_calcium: float | None,
        connection_modifiers: dict,
    ) -> Synapse:
        """Returns a Synapse object."""
        syn_type = cls.determine_synapse_type(syn_description)
        syn_hoc_args = cls.determine_synapse_location(syn_description, cell)

        synapse: Synapse
        if syn_type == SynapseType.GABAAB:
            if condition_parameters.randomize_gaba_rise_time is not None:
                randomize_gaba_risetime = condition_parameters.randomize_gaba_rise_time
            else:
                randomize_gaba_risetime = True
            synapse = GabaabSynapse(cell.cell_id, syn_hoc_args, syn_id, syn_description,
                                    popids, extracellular_calcium, randomize_gaba_risetime)
        elif syn_type == SynapseType.AMPANMDA:
            synapse = AmpanmdaSynapse(cell.cell_id, syn_hoc_args, syn_id, syn_description,
                                      popids, extracellular_calcium)
        else:
            synapse = GluSynapse(cell.cell_id, syn_hoc_args, syn_id, syn_description,
                                 popids, extracellular_calcium)

        synapse = cls.apply_connection_modifiers(connection_modifiers, synapse)

        return synapse

    @staticmethod
    def apply_connection_modifiers(connection_modifiers: dict, synapse: Synapse) -> Synapse:
        if "DelayWeights" in connection_modifiers:
            synapse.delay_weights = connection_modifiers["DelayWeights"]
        if "Weight" in connection_modifiers:
            synapse.weight = connection_modifiers["Weight"]
        if "SynapseConfigure" in connection_modifiers:
            synapse.apply_hoc_configuration(connection_modifiers["SynapseConfigure"])
        return synapse

    @staticmethod
    def determine_synapse_type(
        syn_description: pd.Series,
    ) -> SynapseType:
        """Returns the type of synapse to be created."""
        is_inhibitory: bool = int(syn_description[SynapseProperty.TYPE]) < 100
        all_plasticity_props_available: bool = all(
            x in syn_description and syn_description[x] is not None and not math.isnan(syn_description[x])
            for x in SynapseProperties.plasticity
        )
        no_plasticity_props_available: bool = all(
            x not in syn_description or syn_description[x] is None or math.isnan(syn_description[x])
            for x in SynapseProperties.plasticity
        )
        if is_inhibitory:
            return SynapseType.GABAAB
        else:
            if all_plasticity_props_available:
                return SynapseType.GLUSYNAPSE
            elif no_plasticity_props_available:
                return SynapseType.AMPANMDA
            else:
                raise BluecellulabError("SynapseFactory: Cannot determine synapse type")

    @classmethod
    def determine_synapse_location(cls, syn_description: pd.Series, cell: bluecellulab.Cell) -> SynapseHocArgs:
        """Returns the location of the synapse."""
        isec = int(syn_description[SynapseProperty.POST_SECTION_ID])  # numpy int to int
        section: NeuronSection = cell.get_psection(section_id=isec).hsection

        # old circuits don't have it, it needs to be computed via synlocation_to_segx
        if (SynapseProperty.AFFERENT_SECTION_POS in syn_description and
                not np.isnan(syn_description[SynapseProperty.AFFERENT_SECTION_POS])):
            # position is pre computed in SONATA
            location = syn_description[SynapseProperty.AFFERENT_SECTION_POS]
        else:
            ipt = syn_description[SynapseProperty.POST_SEGMENT_ID]
            syn_offset = syn_description[SynapseProperty.POST_SEGMENT_OFFSET]
            location = cls.synlocation_to_segx(section, ipt, syn_offset)

        return SynapseHocArgs(location, section)

    @staticmethod
    def synlocation_to_segx(
        section: NeuronSection, ipt: float, syn_offset: float
    ) -> float:
        """Translates a synaptic (secid, ipt, offset) to an x coordinate.

        Args:
            section: Section object.
            ipt: Post segment ID.
            syn_offset: Synaptic offset.

        Returns:
            The x coordinate on the section, where the synapse can be placed.
        """
        if syn_offset < 0.0:
            syn_offset = 0.0

        length = section.L

        # access section to compute the distance
        if neuron.h.section_orientation(sec=section) == 1:
            ipt = neuron.h.n3d(sec=section) - 1 - ipt
            syn_offset = -syn_offset

        distance = 0.5
        if ipt < neuron.h.n3d(sec=section):
            distance = (neuron.h.arc3d(ipt, sec=section) + syn_offset) / length
            if distance == 0.0:
                distance = 0.0000001
            if distance >= 1.0:
                distance = 0.9999999

        if neuron.h.section_orientation(sec=section) == 1:
            distance = 1 - distance

        if distance < 0:
            logger.warning(
                f"synlocation_to_segx found negative distance \
                        at curr_sec({neuron.h.secname(sec=section)}) syn_offset: {syn_offset}"
            )
            return 0.0
        else:
            return distance
